use super::db::Noose;
use crate::bussy::{channels, Command, Message, PubSub};
use crate::utils::error::Error;
use crate::utils::structs::Subscription;
use async_trait::async_trait;
use nostr::Event;
use serde::Serialize;
use std::sync::Arc;

use super::user::{User, UserRow};

// Db Interface
pub struct SledDb {
    db: sled::Db,
    events: sled::Tree,
    nip05s: sled::Tree,
    pub users: sled::Tree,

    index: sled::Db,
}

impl SledDb {
    pub fn new() -> Self {
        let db = sled::open("/tmp/sled_db").unwrap();
        let events = db.open_tree("events").unwrap();
        let nip05s = db.open_tree("identifiers").unwrap();
        let accounts = db.open_tree("accounts").unwrap();

        let index = sled::open("/tmp/sled_index").unwrap();

        Self {
            db,
            events,
            nip05s,
            users: accounts,
            index,
        }
    }

    fn clear_db(&self) -> Result<(), sled::Error> {
        self.db.clear()
    }

    fn clear_index(&self) -> Result<(), sled::Error> {
        self.index.clear()
    }

    async fn insert_user(&self, user: UserRow) -> Result<(), Error> {
        let pubkey = user.pubkey.clone();
        let username = user.username.clone();

        if let Ok(Some(_)) = self.nip05s.get(&username) {
            return Err(Error::internal_with_message("User already exists"));
        }

        let mut user_buff = flexbuffers::FlexbufferSerializer::new();
        user.serialize(&mut user_buff).unwrap();

        self.nip05s.insert(&username, user_buff.view()).unwrap();

        let prefix = "nip05:";
        let key = format!("{}{}", prefix, pubkey);
        self.index.insert(key, username.as_bytes()).unwrap();

        Ok(())
    }

    async fn get_user(&self, user: User) -> Result<UserRow, Error> {
        let mut user_row = None;
        if let Some(username) = user.name {
            if let Ok(Some(buff)) = self.nip05s.get(username) {
                let b = flexbuffers::from_slice::<UserRow>(&buff).unwrap();
                user_row = Some(b);
            }
        } else if let Some(pubkey) = user.pubkey {
            let prefix = "nip05:";
            let reference = format!("{}{}", prefix, pubkey);
            if let Ok(Some(row)) = self.index.get(reference) {
                let key = String::from_utf8(row.to_vec()).unwrap();

                if let Ok(Some(buff)) = self.nip05s.get(key) {
                    let b = flexbuffers::from_slice::<UserRow>(&buff).unwrap();

                    user_row = Some(b);
                }
            }
        }
        match user_row {
            Some(user) => Ok(user),
            None => Err(Error::internal_with_message("User not found")),
        }
    }
}

#[async_trait]
impl Noose for SledDb {
    async fn start(&mut self, pubsub: Arc<PubSub>) -> Result<(), Error> {
        let mut subscriber = pubsub.subscribe(channels::MSG_NOOSE).await;

        while let Ok(message) = subscriber.recv().await {
            log::info!("noose subscriber received: {:?}", message);
            let command = match message.content {
                Command::DbReqInsertUser(user) => match self.insert_user(user).await {
                    Ok(_) => Command::DbResOk,
                    Err(e) => Command::ServiceError(e),
                },
                Command::DbReqGetUser(user) => match self.get_user(user).await {
                    Ok(user) => Command::DbResUser(user),
                    Err(e) => Command::ServiceError(e),
                },
                Command::DbReqWriteEvent(event) => match self.write_event(event).await {
                    Ok(_) => Command::DbResOk,
                    Err(e) => Command::ServiceError(e),
                },
                _ => Command::Noop,
            };
            if command != Command::Noop {
                log::info!("Publishing new message");
                let channel = message.source;

                pubsub
                    .publish(
                        channel,
                        Message {
                            source: channels::MSG_NOOSE,
                            content: command,
                        },
                    )
                    .await;
            }
        }

        Ok(())
    }

    async fn migration_up(&self) {}

    async fn write_event(&self, event: Box<Event>) -> Result<String, Error> {
        let mut event_buff = flexbuffers::FlexbufferSerializer::new();
        event.serialize(&mut event_buff).unwrap();

        self.events.insert(event.id, event_buff.view()).unwrap();
        {
            // Timestamp
            let key = format!("created_at:{}|#e:{}", event.created_at, event.id);
            self.index.insert(key, event.id.as_bytes()).unwrap();
        }

        {
            // Author, pubkeys #p
            let key = format!("#author:{}|#e:{}", event.pubkey, event.id);
            self.index.insert(key, event.id.as_bytes()).unwrap();
            // self.index.scan_prefix(
        }

        {
            // Kinds
            let key = format!("#k:{}|#e:{}", event.kind, event.id);
            self.index.insert(key, event.id.as_bytes()).unwrap();
            // self.index.scan_prefix(
        }

        {
            // Tags
            event.tags.iter().for_each(|tag| {
                if let Some(key) = match tag {
                    // #e tag
                    nostr::Tag::Event(event_id, _, _) => Some(format!("#e:{}", event_id)),
                    // #p tag
                    nostr::Tag::PubKey(pubkey, _) => Some(format!("#p:{}|#e:{}", pubkey, event.id)),
                    // #t tag
                    nostr::Tag::Hashtag(hashtag) => Some(format!("#t:{}|#e:{}", hashtag, event.id)),
                    // #a tag
                    nostr::Tag::A {
                        kind,
                        public_key,
                        identifier,
                        relay_url,
                    } => Some(format!(
                        "#a:kind:{}|#a:pubkey:{}#a:identifier:{}|#e:{}",
                        kind, public_key, identifier, event.id
                    )),
                    _ => None,
                } {
                    self.index.insert(key, event.id.as_bytes()).unwrap();
                }
            });

            // let key = format!("#t:{}|#e:{}", event.kind, event.id);
            // self.index.insert(key, event.id.as_bytes()).unwrap();
            // self.index.scan_prefix(
        }

        let message = format!("[\"OK\", \"{}\", true, \"\"]", event.id.to_string());
        Ok(message)
    }

    async fn find_event(&self, subscription: Subscription) -> Result<Vec<String>, Error> {
        todo!()
    }
}

#[cfg(test)]
mod tests {
    use super::SledDb;
    use crate::{
        bussy::PubSub,
        noose::user::{User, UserRow},
    };
    use std::sync::Arc;

    #[tokio::test]
    async fn get_db_names() {
        let pubsub = Arc::new(PubSub::new());

        let db = SledDb::new();

        let pk = "npub1p3ya99jfdafnqlk87p6wfd36d2nme5mkld769rhd9pkht6hmqlaq6mzxdu".to_string();
        let username = "klink".to_string();
        let user = UserRow::new(pk, username, false);
        let result = db.insert_user(user).await;

        let pubkey = "npub1p3ya99jfdafnqlk87p6wfd36d2nme5mkld769rhd9pkht6hmqlaq6mzxdu".to_string();
        let username = "klink".to_string();
        let user = User {
            name: None,
            pubkey: Some(pubkey),
        };
        let user = db.get_user(user).await;

        db.clear_db().unwrap();
        db.clear_index().unwrap();
    }
}
