use super::error::Error;
// use super::nostr_filter_helpers;
use crate::PubSub;
use nostr::{Event, Filter, SubscriptionId};
use std::collections::HashMap;
use std::sync::Arc;
use tokio::sync::mpsc;
use uuid::Uuid;
use warp::ws::Message;

const MAX_SUBSCRIPTIONS: usize = 256;
const MAX_SUBSCRIPTION_ID_LEN: usize = 256;

#[derive(serde::Serialize, PartialEq, Eq, Debug, Clone)]
pub struct Subscription {
    pub id: SubscriptionId,
    pub filters: Vec<Filter>,
}

impl Subscription {
    pub fn new(id: SubscriptionId, filters: Vec<Filter>) -> Self {
        Self { id, filters }
    }

    pub fn get_id(&self) -> String {
        self.id.to_string()
    }

    pub fn needs_historical_events(&self) -> bool {
        self.filters.iter().any(|f| f.limit != Some(0))
    }

    pub fn interested_in_event(&self, event: &Event) -> bool {
        log::info!("[Subscription] Checking if client is interested in the new event");
        for filter in &self.filters {
            if filter.match_event(event) {
                log::info!("[Subscription] found filter that matches the event");
                return true;
            }
        }
        false
    }
}

#[derive(Debug, Clone)]
pub struct Client {
    client_ip_addr: String,
    pub client_id: Uuid,
    pub client_connection: Option<mpsc::UnboundedSender<Result<Message, Error>>>,
    pub subscriptions: HashMap<String, Subscription>,
    max_subs: usize,
}

impl Client {
    pub fn new(client_ip_addr: String) -> Self {
        Self {
            client_ip_addr,
            client_id: Uuid::new_v4(),
            client_connection: None,
            subscriptions: HashMap::new(),
            max_subs: MAX_SUBSCRIPTIONS,
        }
    }

    pub fn ip(&self) -> &str {
        &self.client_ip_addr
    }

    pub fn subscribe(&mut self, subscription: Subscription) -> Result<(), Error> {
        let k = subscription.get_id();
        let sub_id_len = k.len();
        if sub_id_len > MAX_SUBSCRIPTION_ID_LEN {
            log::debug!(
                "Ignoring sub request with excessive length: ({})",
                sub_id_len
            );

            return Err(Error::bad_request(
                "error: subscription request is too long",
            ));
        }

        if self.subscriptions.contains_key(&k) {
            self.subscriptions.remove(&k);
            self.subscriptions.insert(k, subscription.clone());

            log::debug!(
                "replaced existing subscription (cid: {}, sub: {:?})",
                self.client_id,
                subscription.get_id()
            );

            return Ok(());
        }

        if self.subscriptions.len() >= self.max_subs {
            return Err(Error::bad_request(
                "error: max subscriptions limit is exceeded",
            ));
        }

        // Insert subscription
        self.subscriptions.insert(k, subscription);
        log::debug!(
            "registered new subscription, currently have {} active subs (cid: {})",
            self.subscriptions.len(),
            self.client_id
        );
        Ok(())
    }

    pub fn unsubscribe(&mut self, subscription_id: SubscriptionId) {
        self.subscriptions.remove(&subscription_id.to_string());
        log::debug!(
            "removed subscription, currently have {} active subs (cid: {})",
            self.subscriptions.len(),
            self.client_id
        );
    }
}

#[derive(Debug, Clone)]
pub struct Context {
    pub pubsub: Arc<PubSub>,
}

impl Default for Context {
    fn default() -> Self {
        panic!("Use Context::new() to initialize Contexr");
    }
}

impl Context {
    pub fn new() -> Self {
        let pubsub = Arc::new(PubSub::new());

        Self { pubsub }
    }
}
